#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import sys
import traceback
import logging

from lightstreamer.client import LightstreamerClient, Subscription, ClientListener

logger = logging.getLogger(__name__)


class IGStreamService(object):
    def __init__(self, ig_service):
        self.ig_service = ig_service
        self.lightstreamerEndpoint = None
        self.acc_number = None
        self.ls_client = None

    def create_session(self, encryption=False, version="2"):
        ig_session = self.ig_service.create_session(
            encryption=encryption, version=version
        )
        # if we have created a v3 session, we also need the session tokens
        if version == "3":
            self.ig_service.read_session(fetch_session_tokens="true")
        self.lightstreamerEndpoint = ig_session["lightstreamerEndpoint"]
        cst = self.ig_service.session.headers["CST"]
        xsecuritytoken = self.ig_service.session.headers["X-SECURITY-TOKEN"]
        ls_password = "CST-%s|XST-%s" % (cst, xsecuritytoken)

        # Establishing a new connection to Lightstreamer Server
        logger.info("Starting connection with %s" % self.lightstreamerEndpoint)
        self.ls_client = LightstreamerClient(self.lightstreamerEndpoint, None)
        self.ls_client.connectionDetails.setUser(self.acc_number)
        self.ls_client.connectionDetails.setPassword(ls_password)
        try:
            self.ls_client.connect()
            return
        except Exception:
            logger.error("Unable to connect to Lightstreamer Server")
            logger.error(traceback.format_exc())
            sys.exit(1)

    def subscribe(self, subscription: Subscription):
        self.ls_client.subscribe(subscription)

    def unsubscribe(self, subscription: Subscription):
        self.ls_client.unsubscribe(subscription)

    def unsubscribe_all(self):
        # To avoid a RuntimeError: dictionary changed size during iteration
        subscriptions = self.ls_client.getSubscriptions().copy()
        for sub in subscriptions:
            self.ls_client.unsubscribe(sub)

    def add_client_listener(self, listener: ClientListener):
        self.ls_client.addListener(listener)

    def remove_client_listener(self, listener: ClientListener):
        self.ls_client.removeListener(listener)

    def disconnect(self):
        self.unsubscribe_all()
        self.ls_client.disconnect()



from dataclasses import dataclass
from datetime import datetime
from lightstreamer.client import Subscription
from .objects import nan, StreamObject


class TickerSubscription(Subscription):
    """Represents a subscription for tick prices"""

    TICKER_FIELDS = [
        "BID",
        "OFR",
        "LTP",
        "LTV",
        "TTV",
        "UTM",
        "DAY_OPEN_MID",
        "DAY_NET_CHG_MID",
        "DAY_PERC_CHG_MID",
        "DAY_HIGH",
        "DAY_LOW",
    ]

    def __init__(self, epic: str):
        super().__init__(
            mode="DISTINCT",
            items=[f"CHART:{epic}:TICK"],
            fields=self.TICKER_FIELDS,
        )

    def __repr__(self) -> str:
        return f"TickSubscription with {len(self.item_names)} epics"


@dataclass
class Ticker(StreamObject):
    epic: str
    timestamp: datetime = None
    bid: float = nan
    offer: float = nan
    last_traded_price: float = nan
    last_traded_volume: int = 0
    incr_volume: int = 0
    day_open_mid: float = nan
    day_net_change_mid: float = nan
    day_percent_change_mid: float = nan
    day_high: float = nan
    day_low: float = nan

    def __init__(self, epic):
        self.epic = epic

    def __repr__(self) -> str:
        return (
            f"{self.epic}: {self.timestamp.strftime('%Y-%m-%d %H:%M:%S')} "
            f"{self.bid} {self.offer} {self.last_traded_price} "
            f"{self.last_traded_volume} {self.incr_volume} {self.day_open_mid} "
            f"{self.day_net_change_mid} {self.day_percent_change_mid}% {self.day_high} "
            f"{self.day_low}"
        )

    def populate(self, values):
        # print(f"ticker populating: {values}")
        self.set_timestamp_by_name("timestamp", values, "UTM")
        self.set_by_name("bid", values, "BID", float)
        self.set_by_name("offer", values, "OFR", float)
        self.set_by_name("last_traded_price", values, "LTP", float)
        self.set_by_name("last_traded_volume", values, "LTV", int)
        self.set_by_name("incr_volume", values, "TTV", int)
        self.set_by_name("day_open_mid", values, "DAY_OPEN_MID", float)
        self.set_by_name("day_net_change_mid", values, "DAY_NET_CHG_MID", float)
        self.set_by_name("day_percent_change_mid", values, "DAY_PERC_CHG_MID", float)
        self.set_by_name("day_high", values, "DAY_HIGH", float)
        self.set_by_name("day_low", values, "DAY_LOW", float)

    @classmethod
    def identifier(cls, name):
        epic = name.split(":")[1]
        return epic



from datetime import datetime


nan = float("nan")


class StreamObject:
    def set_by_name(self, attr_name, values, key, type):
        try:
            if key in values:
                setattr(self, attr_name, type(values[key]))
        except TypeError:
            # ignore, there will be plenty of dud values
            pass

    def set_timestamp_by_name(self, attr_name, values, key):
        try:
            if key in values:
                setattr(
                    self, attr_name, datetime.fromtimestamp(int(values[key]) / 1000)
                )
        except TypeError:
            # ignore, there will be plenty of dud values
            pass



import logging
from queue import Queue
from threading import Thread
import time

from lightstreamer.client import SubscriptionListener, ItemUpdate

from trading_ig import IGStreamService
from .ticker import Ticker
from .ticker import TickerSubscription

logger = logging.getLogger(__name__)


class StreamingManager:
    def __init__(self, service: IGStreamService):
        self._service = service
        self._subs = {}

        # setup data objects
        self._tickers = {}

        # set up consumer queue
        self._queue = Queue()
        self._consumer_thread = Consumer(self._queue, self)
        self._consumer_thread.start()

    @property
    def service(self):
        return self._service

    @property
    def tickers(self):
        return self._tickers

    def start_tick_subscription(self, epic) -> TickerSubscription:
        tick_sub = TickerSubscription(epic)
        tick_sub.addListener(TickerListener(self._queue))
        self.service.subscribe(tick_sub)
        self._subs[epic] = tick_sub
        return tick_sub

    def stop_tick_subscription(self, epic):
        subscription = self._subs.pop(epic)
        self.service.unsubscribe(subscription)

    def ticker(self, epic, timeout_length=3):
        # we won't have a ticker until at least one update is received from server,
        # let's give it a few seconds
        timeout = time.time() + timeout_length
        while True:
            logger.info(f"Waiting for ticker for '{epic}'...")
            if epic in self._tickers or time.time() > timeout:
                break
            time.sleep(0.25)
        try:
            ticker = self._tickers[epic]
        except KeyError:
            raise Exception(
                f"No ticker found for {epic} after "
                f"waiting {timeout_length} seconds - giving up"
            )
        return ticker

    def on_update(self, update):
        self._queue.put(update)

    def stop_subscriptions(self):
        logger.info("Unsubscribing from all")
        self.service.unsubscribe_all()
        self.service.disconnect()
        if self._consumer_thread:
            self._consumer_thread.join(timeout=5)
            self._consumer_thread = None


class TickerListener(SubscriptionListener):
    def __init__(self, queue: Queue) -> None:
        self._queue = queue

    def onItemUpdate(self, update: ItemUpdate):
        self._queue.put(update)

    def onSubscription(self):
        logger.info("TickerListener onSubscription()")

    def onSubscriptionError(self, code, message):
        logger.info(f"TickerListener onSubscriptionError(): '{code}' {message}")

    def onUnsubscription(self):
        logger.info("TickerListener onUnsubscription()")


class Consumer(Thread):
    def __init__(self, queue: Queue, manager: StreamingManager):
        super().__init__(name="ConsumerThread", daemon=True)
        self._queue = queue
        self._manager = manager

    @property
    def manager(self):
        return self._manager

    def run(self):
        logger.info("Consumer: Running")
        while True:
            item = self._queue.get()

            # deal with each different type of update
            name = item.getItemName()
            if name.startswith("CHART:"):
                self._handle_ticker_update(item)

            logger.debug(f"Consumer thread alive. queue length: {self._queue.qsize()}")

    def _handle_ticker_update(self, item: ItemUpdate):
        epic = Ticker.identifier(item.getItemName())

        if epic not in self.manager.tickers:
            ticker = Ticker(epic)
            self.manager.tickers[epic] = ticker

        ticker = self.manager.tickers[epic]
        ticker.populate(item.getChangedFields())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
IG Markets API Library for Python
https://github.com/ig-python/trading-ig/
by Femto Trader - https://github.com/femtotrader
"""

from __future__ import absolute_import, division, print_function


from .rest import IGService
from .stream import IGStreamService

__all__ = [
    "IGService",
    "IGStreamService",
]
